from py_cc_ohlcv import exchanges

# instrument_mapping: Apply this to the scraper's instrument symbol 
# timestamp_format: String or Int representing second multiplier operand (e.g. 1000 * seconds = milliseconds)
# response_keys: List of keys to traverse down a nested JSON structure received in the request response

exchange_metadata = {
    exchanges.BINANCE : {
        'instrument_mapping' : (lambda inst: ''.join(inst.split("_"))),
        'url' : "https://api.binance.com/api/v3/klines?symbol={}&interval={}&startTime={}&endTime={}&limit=1000",
        'timestamp_format' : 1000,
        'resolutions': {
            '1m': '1m',
            '5m': '5m',
            '15m': '15m',
            '1h': '1h',
            '6h': '6h',
            '1d': '1d'
        },
        'max_candles_per_req': 1000,
        'response_keys' : []
    },
    exchanges.BINANCE_US: {
        'instrument_mapping' : (lambda inst: ''.join(inst.split("_"))),
        'url' : "https://api.binance.us/api/v3/klines?symbol={}&interval={}&startTime={}&endTime={}&limit=1000",
        'timestamp_format' : 1000,
        'resolutions': {
            '1m': '1m',
            '5m': '5m',
            '15m': '15m',
            '1h': '1h',
            '6h': '6h',
            '1d': '1d'
        },
        'max_candles_per_req': 1000,
        'response_keys' : []
    },
    exchanges.BINANCE_FUTURES: {
        'instrument_mapping' : (lambda inst: ''.join(inst.split("_"))),
        'url' : "https://fapi.binance.com/fapi/v1/klines?symbol={}&interval={}&startTime={}&endTime={}&limit=1500",
        'timestamp_format' : 1000,
        'resolutions': {
            '1m': '1m',
            '5m': '5m',
            '15m': '15m',
            '1h': '1h',
            '6h': '6h',
            '1d': '1d'
        },
        'max_candles_per_req': 1000,
        'response_keys' : []
    },
    exchanges.COINBASE_PRO: {
        'instrument_mapping' : (lambda inst: '-'.join(inst.split("_"))),
        'url' : "https://api.exchange.coinbase.com/products/{}/candles?granularity={}&start={}&end={}",
        'timestamp_format' : "%Y-%m-%dT%H:%M:%S.%fZ",
        'resolutions': {
            '1m': 60,
            '5m': 300,
            '15m': 900,
            '1h': 3600,
            '6h': 21600,
            '1d': 86400
        },
        'max_candles_per_req': 300,
        'response_keys': []
    },
    exchanges.FTX: {
        'instrument_mapping' : (lambda inst: '/'.join(inst.split("_"))),
        'url' : "https://ftx.com/api/markets/{}/candles?resolution={}&start_time={}&end_time={}",
        'timestamp_format' : 1,
        'resolutions' : {
            '1m': 60
        },
        'max_candles_per_req': 300,
        'response_keys': ['result']
    },
    exchanges.GEMINI: {
        'instrument_mapping' : (lambda inst: ''.join(inst.split("_"))),
        'url' : "https://api.gemini.com/v2/candles/{}/{}",
        'timestamp_format' : 1000,
        'resolutions' : {
            '1m': '1m'
        },
        'max_candles_per_req': 300,
        'response_keys': []
    },
    exchanges.KUCOIN: {
        'instrument_mapping' : (lambda inst: '-'.join(inst.split("_"))),
        'url' : "https://api.kucoin.com/api/v1/market/candles?symbol={}&type={}&startAt={}&endAt={}",
        'timestamp_format' : 1,
        'resolutions': {
            '1m': '1min'
        },
        'max_candles_per_req': 300,
        'response_keys': ['data']
    }
}


from py_cc_ohlcv import exchanges, exchange_metadata, exchange_functions
import datetime
import logging
import requests
import time

import pandas as pd

logger = logging.getLogger()

class Scraper:
    exchange = None
    metadata = None

    instrument = None
    symbol = None
    resolution = None
    resolution_val = None

    url = None
    proxies = None
    window_start = None
    window_end = None

    def __init__(self, exchange, instrument, resolution, start_time: datetime, end_time: datetime):
        self.exchange = exchange
        self.instrument = instrument

        self.metadata = exchange_metadata.exchange_metadata[self.exchange]
        self.exchange_functions = exchange_functions.exchange_functions
        self.symbol = exchange_metadata.exchange_metadata[self.exchange]['instrument_mapping'](instrument)
        self.url = exchange_metadata.exchange_metadata[self.exchange]['url']

        self.resolution = resolution
        self.resolution_val = exchange_metadata.exchange_metadata[self.exchange]['resolutions'][self.resolution]

        self.window_start = start_time
        self.window_end = end_time

    def set_proxies(self, proxies: dict):
        self.proxies = proxies

    def run(self):
        logger.info("""Executing job with parameters: \n 
        [EXCHANGE] {}, [INSTR] {}, [RESOLUTION] {} \n
        [START] {} ({})\n
        [END] {} ({})
        """.format(self.exchange, self.symbol, self.resolution,
        self.window_start, self.window_start.timestamp() * 1000,
        self.window_end, self.window_end.timestamp() * 1000))

        # Set time delta based on maximum candles per request and desired resolution
        delta_t = None 
        if self.resolution == '1m':
            delta_t = datetime.timedelta(minutes=self.metadata['max_candles_per_req'])
        else:
            logger.error("Resolution not supported!")
            return
        
        start = self.window_start
        end = self.window_end
        timestamp_format = self.metadata['timestamp_format']
        parsed_candles = []

        # Exception for coinbase
        if self.exchange == exchanges.COINBASE_PRO:
            start = start - delta_t / self.metadata['max_candles_per_req']

        while start < self.window_end:
            end = start + delta_t
            if (end > self.window_end):
                end = self.window_end

            url = self.url.format(
                self.symbol,
                self.resolution_val,
                start.strftime(timestamp_format) if type(timestamp_format) is str else int(start.timestamp()) * timestamp_format,
                end.strftime(timestamp_format) if type(timestamp_format) is str else int(end.timestamp()) * timestamp_format
            )

            logger.debug(url)
            response = requests.get(url=url, proxies=self.proxies)
            logger.debug(response.headers)

            candles_response = response.json()
            for key in self.metadata['response_keys']:
                candles_response = candles_response[key]

            logger.debug("Received {} candles".format(len(candles_response)))

            for candle in candles_response:
                parsed_candles.append(self.exchange_functions[self.exchange +'_candle_mapping'](candle))

            time.sleep(0.25) # TODO: Use an actual rate limiter
            start = end
        
        df = pd.DataFrame(parsed_candles)
        df = df.set_index('open_timestamp')
        df = df.sort_index(ascending=True)
        df = df[~df.index.duplicated(keep='first')] # Drop duplicate timestamps
        df = df.truncate(after=int(self.window_end.timestamp() * 1000))

        return df


import datetime as DT

exchange_functions = {}

def register(func):
    exchange_functions[func.__name__] = func
    return func

@register
def binance_candle_mapping(candle):
    entry = {
        "open_timestamp" : int(candle[0]),
        "open" : float(candle[1]),
        "high" : float(candle[2]),
        "low" : float(candle[3]),
        "close" : float(candle[4]),
        "volume" : float(candle[5])
    }
    return entry

@register
def binance_us_candle_mapping(candle):
    entry = {
        "open_timestamp" : int(candle[0]),
        "open" : float(candle[1]),
        "high" : float(candle[2]),
        "low" : float(candle[3]),
        "close" : float(candle[4]),
        "volume" : float(candle[5])
    }
    return entry

@register
def coinbase_pro_candle_mapping(candle):
    entry = {
        "open_timestamp" : int(candle[0]) * 1000,
        "open" : candle[3],
        "high" : candle[2],
        "low" : candle[1],
        "close" : candle[4],
        "volume" : candle[5]
    }
    return entry

@register 
def ftx_candle_mapping(candle):
    entry = {
        "open_timestamp" : int(DT.datetime.strptime(candle['startTime'], "%Y-%m-%dT%H:%M:%S+00:00").replace(tzinfo=DT.timezone.utc).timestamp() * 1000),
        "open" : candle['open'],
        "high" : candle['high'],
        "low" : candle['low'],
        "close" : candle['close'],
        "volume" : candle['volume']
    }
    return entry

@register
def gemini_candle_mapping(candle):
    entry = {
        "open_timestamp" : candle[0],
        "open" : candle[1],
        "high" : candle[2],
        "low" : candle[3],
        "close" : candle[4],
        "volume" : candle[5]
    }
    return entry

@register
def kucoin_candle_mapping(candle):
    entry = {
        "open_timestamp" : int(DT.datetime.utcfromtimestamp(int(candle[0])).replace(tzinfo=DT.timezone.utc).timestamp() * 1000),
        "open" : float(candle[1]),
        "high" : float(candle[3]),
        "low" : float(candle[4]),
        "close" : float(candle[2]),
        "volume" : float(candle[5])
    }
    return entry


from setuptools import setup, find_packages

setup(
    name='py_cc_ohlcv',
    version='0.1.1',
    license='MIT',
    author="Max Gao",
    author_email='gaamox@tutanota.com',
    packages=find_packages(exclude=["examples"]),
    url='https://github.com/gmax0/Cryptocurrency-OHLCV-Scraper',
    keywords='ohlcv, cryptocurrency, historical, market, data',
    install_requires=[
        'pandas',
        'requests'
      ],
)
